from datetime import datetime, timedelta

import pandas as pd
import logging


def _remove_similar_consecutive_values_single_monitor(data_frame, epsilon):
    if data_frame.size > 0:
        to_remove = []

        pivot = data_frame.iloc[0]
        for idx, row in data_frame.iloc[1:].items():
            if abs(pivot - row) <= epsilon:
                to_remove.append(idx)
            else:
                pivot = row

        data_frame.drop(to_remove, axis=0, inplace=True)

    return data_frame

def _remove_similar_consecutive_values(data_frame, monitor_name, epsilon):
    if data_frame.size > 0:
        to_remove = []

        pivot = data_frame.loc[0][1:]
        for idx, row in data_frame.iterrows():
            if idx == 0:
                continue
            if ((pivot - row[1:]) ** 2).sum() ** 0.5 <= epsilon:
                to_remove.append(idx)
            else:
                pivot = row

        data_frame.drop(to_remove, axis=0, inplace=True)

    return data_frame


def _convert(data_frame):
    data_frame['TimeStampLong'] = pd.to_datetime(data_frame['TimeStampLong'], unit='us')

    return data_frame


def _merge_data_frames(data_frames):
    data_frame = data_frames[0]
    if len(data_frames) >= 2:
        data_frame = pd.merge(data_frames[0], data_frames[1], how='outer')
        for idx in range(2, len(data_frames)):
            data_frame = pd.merge(data_frame, data_frames[idx], how='outer')

    # data_frame['TimeStampLong'] = pd.to_datetime(data_frame['TimeStampLong'], unit='us')
    data_frame.sort_values(by=['TimeStampLong'], inplace=True)

    return data_frame


def _filter(data_frame, q_data_ini, q_data_end):
    data_frame = data_frame[
        (data_frame['TimeStampLong'] >= q_data_ini) &
        (data_frame['TimeStampLong'] < q_data_end)]

    return data_frame



import os
import io

import pandas as pd
import logging

from mmwebreport.core.api import Report
from mmwebreport.retrieve import cache
from mmwebreport.retrieve.processing import _remove_similar_consecutive_values, _convert, _merge_data_frames, _filter


class RetrieveMonitor(object):
    """
    A class used to make a high level request to the Monitor Manager Web Report, with the following functionality:

        - Simple interface to extract with the minimum requery information complex data.
        - Cache management to optimize the time necessary to obtain the data.
        - Pagination of the result to treat with large volume of data.
        - Filtering similar values base on epsilon value.
        - Treatment of all monitor data type in a uniform way: n-dimension monitor and enumerates.
        - Integration with Pandas to generate Data Frame to easily do data science with tabular data.
        - Integrity sanity check of the request.
    """
    def __init__(self, host, port, q_query, q_name, q_clean_cache=False, q_fillfw=False):
        self.request = Report(host, port)
        self._query = q_query
        self._query_name = q_name
        self._path = os.path.expanduser("~") + "/.cache/webreport/monitormanager"
        self._clean_cache = q_clean_cache
        self._fillfw = q_fillfw

    def retrieve_raw(self, date_ini, date_end, q_entry):
        """
        Given a date interval and a query entry, it creates a data frame with the result. The request result
        is cached on disk.

        :param date_ini: The search request initial date and time.
        :param date_end: The search request end date and time.
        :param q_entry: The search request query entry. For example:
            {
                "component": "MACS.AzimuthAxis",
                "monitor": "position",
                "epsilon": 0.5,
                "type": "monitor"
            }

        :return: a data frame for this request.
        """
        # Create a cache object
        a_cache = cache.CacheRaw(self._path, date_ini, date_end, self._query_name, [q_entry])
        data_frames_page = []
        # Iterate over the result pages.
        cursor = self.request.search(date_ini, date_end, [q_entry])
        for page, c in enumerate(cursor):
            logging.info("Retrieve monitor from %s to %s page %s", date_ini, date_end, page)
            if not self._clean_cache and a_cache.is_data_frame_cached(page):
                logging.info("Retrieve raw monitor from cache")
                data_frame = a_cache.data_frame_from_cache(page)
            else:
                logging.info("Retrieve monitor from API call")
                data_frame = pd.read_csv(io.StringIO(c.run().to_csv()), sep=",")
                a_cache.data_frame_to_cache(data_frame, page)
            data_frames_page.append(data_frame)
        if len(data_frames_page) != 0:
            data_frame = pd.concat(data_frames_page, ignore_index=True, sort=False)
        else:
            data_frame = pd.DataFrame([], columns=['TimeStampLong'])

        return data_frame

    def retrieve_filtered(self, time_interval_ini, time_interval_end, date_ini, date_end, q_entry):
        """
        Given a time interval, date interval and a query entry, it creates a data frame with the result, filtered by
        similar values.

        The request to the Monitor Manager Web report is done using the time interval and cached on disk as raw data.
        The data frame returned will be inside the date range and also cached. Normally, the time interval will have
        a fixed size, making easier to reuse the data raw cached between different queries.

        :param time_interval_end: The search request initial date and time.
        :param time_interval_ini: The search request end date and time.
        :param date_ini: The search filtered request initial date and time.
        :param date_end: The search filtered  request end date and time.
        :param q_entry: The search request query entry.
            {
                "component": "MACS.AzimuthAxis",
                "monitor": "position",
                "epsilon": 0.5,
                "type": "monitor"
            }

        :return: a data frame for this request, filtering the similar values base on epsilon monitor param.
        """
        logging.info("Retrieve monitor from %s to %s ", date_ini, date_end)
        # Create a cache object
        a_cache = cache.CacheFiltered(self._path, date_ini, date_end, self._query_name, [q_entry])
        if not self._clean_cache and a_cache.is_data_frame_cached():
            logging.info("Retrieve raw monitor from cache")
            data_frame = a_cache.data_frame_from_cache()
        else:
            logging.info("Retrieve monitor from API call")
            data_frame = self.retrieve_raw(time_interval_ini, time_interval_end, q_entry)
            # Convert time stamp to date
            data_frame = _convert(data_frame)
            # Filter data inside time interval
            data_frame = _filter(data_frame, date_ini, date_end)
            data_frame.reset_index(drop=True, inplace=True)
            # Remove similar values using the epsilon to compare if abs(a-b) < epsilon them a == b
            if q_entry["type"] == "monitor" or q_entry["type"] == "array":
                data_frame = _remove_similar_consecutive_values(data_frame,
                                                                q_entry['component'] + "." + q_entry['monitor'],
                                                                q_entry["epsilon"])
            a_cache.data_frame_to_cache(data_frame)

        return data_frame

    def retrieve_summary_chunk(self, time_interval_ini, time_interval_end, q_date_ini, q_date_end):
        """
        Given a time interval, date interval and a query, it creates a data frame with the result, containing
        several monitors, each one filtered by similar values.

        The request to the Monitor Manager Web report is done using the time interval and cached on disk as raw data
        for each monitor. Then, the filtered samples for each monitor in the date interval is obtained from raw data,
        and also cached. Finally, a data frame will be returned with samples for each monitor, inside the data range.
        This summary will be also cached.

        :param time_interval_ini: The search request initial date and time.
        :param time_interval_end: The search request end date and time.
        :param q_date_ini: The search filtered request initial date and time.
        :param q_date_end: The search filtered  request end date and time.

        :return: a data frame for this request, filtering the similar values base on epsilon monitor param and grouping
        several monitor.
        """
        logging.info("Retrieve hourly from %s to %s", time_interval_ini, time_interval_end)

        # Create a cache object
        a_cache = cache.CacheSummaryHourly(self._path,
                                           time_interval_ini, time_interval_end,
                                           self._query_name, self._query)

        if not self._clean_cache and a_cache.is_data_frame_cached():
            logging.info("Retrieve summary hourly from cache")
            data_frame = a_cache.data_frame_from_cache()
        else:
            logging.info("Retrieve summary hourly from API call")
            # Iterate over the query entries in the query
            data_frames_hourly = []
            for idx, monitor in enumerate(self._query):
                logging.info("Retrieve %s %s %s ", time_interval_ini, time_interval_end, monitor)
                data_frame = self.retrieve_filtered(time_interval_ini, time_interval_end, q_date_ini, q_date_end,
                                                    monitor)
                data_frames_hourly.append(data_frame)
            logging.info("Merge hours %s %s ", q_date_ini, q_date_end)
            # Merge all the results
            data_frame = _merge_data_frames(data_frames_hourly)
            a_cache.data_frame_to_cache(data_frame)

        return data_frame

    def retrieve_summary(self, date_range):
        """
        Given date range it creates a data frame with the result, containing several monitors, each one filtered by
        similar values.

        It is important to notice than both date and time are interpreted as intervals and not as initial and end
        datetime.

        For example, if the input is:

            date_ini: 2022-07-01 19:00:00
            date_end: 2022-07-31 06:30:00

        The result will include all the sample between 2022 07 01 and 2022 07 31, where the time is between 19:00 and
        06:30:00 and not samples from 2022-07-01 19:00:00 to 2022-07-31 06:00:00.

        In addition, in order to optimize the time necessary to obtain the data, a different cache levels are used: raw
        data from monitor manager web report (this raw data is fixed by individual monitor and with a default chunk size
        of one hour), filtered data between the date range where similar values are excluded, summary data collapsing all
        the monitors filtered in the previous step , and the final data for the date whole range.

        The query is defined using a dictionary object as follow:

            query =
                [
                    {
                        "component": "",
                        "monitor": "",
                        "epsilon": ,
                        "type": ""
                    }
                ]

        Where:

            component:  The component name following GCS convention but replacing slash with dot. For example,
                        MACS.AzimuthAxis.

            monitor:    The monitor name following GCS convention. For example: position.

            epsilon:    If different from 0, a value used to filter consecutive similar samples from a query resutl.

            type:       The typ of monitor. It can be monitor for one-dimensional monitor, array for n-dimensional
                        monitor and magnitude for monitor of type enumerated.


        In order to clarify how this works let see and example:

        Given:

            date_ini: 2022-07-01 19:00:00
            date_end: 2022-07-01 19:00:10
            query_name: following_error_study
            query = \
                [
                    {
                        "component": "OE.ObservingEngine",
                        "monitor": "currentObservingState",
                        "epsilon": 0,
                        "type": "magnitude"
                    },
                    {
                        "component": "MACS.AzimuthAxis",
                        "monitor": "position",
                        "epsilon": 0.5,
                        "type": "monitor"
                    }
                ]

        The following structure will be created on disk:

        .
        ├── 2022-07-01
        │ └── 19
        │     ├── MACS
        │     │ └── AzimuthAxis
        │     │     └── position
        │     │         ├── 2022-07-01_19_00_00.2022-07-01_19_00_10.MACS.AzimuthAxis.position.csv.gz
        │     │         ├── 2022-07-01_19_00_00.2022-07-01_19_00_10.MACS.AzimuthAxis.position.json
        │     │         └── raw
        │     │             ├── 2022-07-01_19_00_00.2022-07-01_20_00_00.MACS.AzimuthAxis.position.raw.0000.csv.gz
        │     │             └── 2022-07-01_19_00_00.2022-07-01_20_00_00.MACS.AzimuthAxis.position.raw.0000.json
        │     ├── OE
        │     │ └── ObservingEngine
        │     │     └── currentObservingState
        │     │         ├── 2022-07-01_19_00_00.2022-07-01_19_00_10.OE.ObservingEngine.currentObservingState.csv.gz
        │     │         ├── 2022-07-01_19_00_00.2022-07-01_19_00_10.OE.ObservingEngine.currentObservingState.json
        │     │         └── raw
        │     │             ├── 2022-07-01_19_00_00.2022-07-01_20_00_00.OE.ObservingEngine.currentObservingState.raw.0000.csv.gz
        │     │             └── 2022-07-01_19_00_00.2022-07-01_20_00_00.OE.ObservingEngine.currentObservingState.raw.0000.json
        │     └── summary
        │         ├── 2022-07-01_19_00_00.2022-07-01_20_00_00.following_error_study.summary.cvs.gz
        │         └── 2022-07-01_19_00_00.2022-07-01_20_00_00.following_error_study.summary.json
        └── summary
            └── following_error_study
                ├── 2022-07-01_19_00_00.2022-07-01_19_00_10.following_error_study.summary.cvs.gz
                └── 2022-07-01_19_00_00.2022-07-01_19_00_10.following_error_study.summary.json

        In the 2022-07-01/19/MACS/AzimuthAxis/position/raw/
            2022-07-01_19_00_00.2022-07-01_20_00_00.MACS.AzimuthAxis.position.raw.0000.csv.gz file are the
        samples for one monitor where time interval chunk has been fixed to one hour inside the date interval requested
        and with one file per page.

        In the 2022-07-01/19/MACS/AzimuthAxis/position/
            2022-07-01_19_00_00.2022-07-01_20_00_00.MACS.AzimuthAxis.position.csv.gz file are the
        samples for one monitor between the time interval chunk filtered by date interval and similar values.

        In the 2022-07-01/19/summary/
            2022-07-01_19_00_00.2022-07-01_20_00_00.following_error_study.summary.cvs.gz are the
        samples for all monitors between the time interval chunk filtered by date interval and similar values.

        In the 2022-07-01/summary/
            2022-07-01_19_00_00.2022-07-01_19_00_10.following_error_study.summary.cvs.gz are all the
        samples for all monitors filtered by date interval and similar values.

        :param date_range: A data range strategy. Range can be by date, date range, week, month, year.

        :return: a data frame filtering the similar values base on epsilon monitor param and grouping several monitor.
        """
        logging.info("Retrieve Summary from with range %s", date_range)

        # Create a cache object
        a_cache = cache.CacheSummary(self._path,
                                     date_range.get_date_init(), date_range.get_date_end(),
                                     self._query_name, self._query)

        if not self._clean_cache and a_cache.is_data_frame_cached():
            logging.info("Retrieve summary hourly from cache")
            data_frame = a_cache.data_frame_from_cache()
        else:
            logging.info("Retrieve summary hourly from API call")
            data_frames_monitors = []
            time_intervals = date_range.make_interval()
            for time_interval in time_intervals:
                logging.info("Retrieve hourly from %s to %s", time_interval[0], time_interval[1])
                data_frame = self.retrieve_summary_chunk(time_interval[0], time_interval[1],
                                                         time_interval[2], time_interval[3])
                data_frames_monitors.append(data_frame)
            data_frame = _merge_data_frames(data_frames_monitors)
            a_cache.data_frame_to_cache(data_frame)
        # todo cache this configuration
        if self._fillfw:
            data_frame.fillna(method='ffill', inplace=True)
            data_frame.dropna(inplace=True)

        return data_frame

    def sanity_query_check(self, a_query):
        """
        Check if a query is well-formed:

            - Component, monitor, epsilon and type have been specified.
            - Epsilon is int of float type.
            - Type is one of the defined values: monitor, array or magnitude
            - The monitor is defined in the system and the type is coherent

        :param a_query: a query with the folling structure
            {
                "component": "MACS.AzimuthAxis",
                "monitor": "position",
                "epsilon": 0.5,
                "type": "monitor"
            }
        :return: exception in case the query is not well-formed.
        """
        if not isinstance(a_query, (list, tuple)):
            raise "A query must be a list of monitors."
        else:
            for e in a_query:
                if not "component" in e: raise "Component not specified."
                if not "monitor" in e: raise "Monitor not specified."
                if not "epsilon" in e: raise "Epsilon not specified."
                if not "type" in e: raise "Type not specified."

                if not (type(e['epsilon']) == float or type(e['epsilon']) == int): raise "Epsilon should be float"
                if not e['type'] in ["monitor", "array", "magnitude"]:
                    raise "Not valid value for type should be: monitor, array or enum"

                try:
                    config = self.request.get_monitor_configuration(q_component=e['component'].replace('.', '/'),
                                                                    q_monitor=e['monitor'],
                                                                    q_type=e['type'])

                    if e['type'] in "array" and config["dimension_x"] == 1 and config["dimension_y"] == 1:
                        raise "Monitor dimension incorrect"
                except:
                    raise "Monitor does not exist in database."


from datetime import datetime
import pandas as pd

data_time_format = "%Y-%m-%d %H:%M:%S"
week_time_format = "%G-W%V-%u"
month_time_format = "%Y-%m"
year_time_format = "%Y"
time_format = "%H:%M:%S"


# todo include docs


class DateRange(object):

    def __init__(self, date_format=data_time_format):
        self._date_ini = None
        self._date_end = None
        self._frequency = None
        self._time_interval = None
        self._intervals = []
        self._time_ini = None
        self._time_end = None
        self._date_format = date_format
        # In case time end is less than time ini, increase the date.
        self._date_end_plus_one_day = None
        # Setup cache step
        self._date_cache_step = None

    def get_date_init(self):
        return pd.Timestamp.combine(self._date_ini, self._time_ini.time())

    def get_date_end(self):
        return pd.Timestamp.combine(self._date_end + self._date_end_plus_one_day, self._time_end.time())

    def _parse_frequency_cache(self):
        self._date_cache_step = pd.offsets.Hour(1)
        if self._frequency in '1H':
            self._date_cache_step = pd.offsets.Hour(1)
        elif self._frequency in '30min':
            self._date_cache_step = pd.offsets.Minute(30)

    def _parse_date(self):
        # Parse date, for the moment date ini and date end are the same
        self._date_ini = pd.to_datetime(self._date_ini, format=self._date_format)
        self._date_end = pd.to_datetime(self._date_end, format=self._date_format)

    def _parse_time_interval(self):
        self._date_end_plus_one_day = pd.DateOffset(0)

        if self._time_interval:
            if self._time_interval[0] > self._time_interval[1]:
                self._date_end_plus_one_day = pd.DateOffset(1)

            self._time_ini = pd.to_datetime(self._time_interval[0], format=time_format)
            self._time_end = pd.to_datetime(self._time_interval[1], format=time_format)

        else:
            self._date_end_plus_one_day = pd.DateOffset(1)

            self._time_ini = pd.Timestamp(
                datetime(self._date_ini.year, self._date_ini.month, self._date_ini.day, 0, 0, 0))
            self._time_end = pd.Timestamp(
                datetime(self._date_ini.year, self._date_ini.month, self._date_ini.day, 0, 0, 0))

    def _make_time_intervals_by_date_core(self, date_ini, time_ini, date_end, time_end):
        # Round init time to near chunk frequency.
        freq_date_ini = pd.Timestamp.combine(date_ini.date(), time_ini.floor(self._frequency).time())
        # Ceil end time to near chunk frequency
        freq_date_end = pd.Timestamp.combine(date_end.date(), time_end.ceil(self._frequency).time())

        date_time_ini = pd.Timestamp.combine(date_ini.date(), time_ini.time())
        date_time_end = pd.Timestamp.combine(date_end.date(), time_end.time())

        intervals = []
        pivot_date = freq_date_ini
        while pivot_date < freq_date_end:
            date_ini_for_cache = pivot_date
            date_end_for_cache = pivot_date + self._date_cache_step

            data_ini_interval = date_ini_for_cache if date_ini_for_cache > date_time_ini else date_time_ini
            date_end_interval = date_end_for_cache if date_end_for_cache < date_time_end else date_time_end

            intervals.append((date_ini_for_cache, date_end_for_cache,
                              data_ini_interval, date_end_interval))

            pivot_date += self._date_cache_step

        return intervals

    def make_interval(self):
        date_pivot = self._date_ini
        while date_pivot < self._date_end + self._date_end_plus_one_day:
            self._intervals += self._make_time_intervals_by_date_core(date_pivot,
                                                                      self._time_ini,
                                                                      date_pivot + self._date_end_plus_one_day,
                                                                      self._time_end)
            date_pivot += pd.DateOffset(1)

        return self._intervals

    def build(self, date_interval, frequency="1H", time_interval=None):
        self._date_ini = date_interval[0]
        self._date_end = date_interval[1 if len(date_interval) == 2 else 0]
        self._frequency = frequency
        self._time_interval = time_interval

        self._parse_date()
        self._parse_time_interval()
        self._parse_frequency_cache()

    def from_json(self, query_json):
        self.build(query_json["date_interval"],
                   query_json["frequency"],
                   query_json["time_interval"] if "time_interval" in query_json else [])

    def __str__(self):
        return str(self.get_date_init()) + " - " + str(self.get_date_end())


class DateRangeByDate(DateRange):
    def __init__(self):
        super().__init__()


class DateRangeByDateRange(DateRange):
    def __init__(self):
        super().__init__()


class DateRangeByWeek(DateRange):
    def __init__(self):
        super().__init__(week_time_format)

    def _parse_date(self):
        # Parse date, for the moment date ini and date end are the same
        self._date_ini = pd.to_datetime(self._date_ini + "-1", format=self._date_format)
        self._date_end = pd.to_datetime(self._date_end + "-1", format=self._date_format) + pd.DateOffset(days=7 - 1)


class DateRangeByMonth(DateRange):
    def __init__(self):
        super().__init__(month_time_format)

    def _parse_date(self):
        # Parse date, for the moment date ini and date end are the same
        self._date_ini = pd.to_datetime(self._date_ini + "-1", format=self._date_format)
        self._date_end = pd.to_datetime(self._date_end + "-1", format=self._date_format) + \
                         pd.DateOffset(days=self._date_ini.days_in_month - 1)


class DateRangeByYear(DateRange):
    def __init__(self):
        super().__init__(year_time_format)

    def _parse_date(self):
        # Parse date, for the moment date ini and date end are the same
        self._date_ini = pd.to_datetime(self._date_ini, format=self._date_format)
        self._date_end = pd.to_datetime(self._date_end, format=self._date_format) + \
                         pd.DateOffset(days=365 + self._date_ini.is_leap_year - 1)



import argparse
import json
import logging
import sys

_DEFAULT_HOST_NAME = "calp-vwebrepo"
_DEFAULT_HOST_PORT = "8081"

from mmwebreport.retrieve.daterange import *
from mmwebreport.retrieve.retrieve import RetrieveMonitor

def Main():
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument('-q', '--query_file',
                        required=True,
                        type=str,
                        default="query.json",
                        dest="query_file",
                        help="Query configuration file")
    parser.add_argument('--host',
                        required=False,
                        type=str,
                        default=_DEFAULT_HOST_NAME,
                        dest="host_name",
                        metavar="<monitor manager web report server host>",
                        help="Monitor Manager Web Report Server host name")
    parser.add_argument('-p', '--port',
                        required=False,
                        type=str,
                        default=_DEFAULT_HOST_PORT,
                        dest="port_name",
                        help="Monitor Manager Web Report Server port number")
    parser.add_argument('-c', '--clean_cache',
                        required=False,
                        type=bool,
                        default=False,
                        dest="clean_cache",
                        help="Clean cache of previous Monitor Manager Reports")

    args = parser.parse_args()

    with open(args.query_file) as json_file:
        json_query = json.load(json_file)

    query_name = json_query["query_config"]["query_name"]
    query_fill_forwarder = json_query["query_config"]["query_fill_forwarder"]

    query_monitor_list = json_query["query_monitor_list"]

    retrieve = RetrieveMonitor(_DEFAULT_HOST_NAME, _DEFAULT_HOST_PORT,
                               query_monitor_list,
                               query_name,
                               q_clean_cache=args.clean_cache,
                               q_fillfw=query_fill_forwarder)

    date_range_list = json_query["query_date_range"]

    DateRange = getattr(sys.modules[__name__], date_range_list["type"])
    date_range = DateRange()
    date_range.from_json(date_range_list)

    data_frame = retrieve.retrieve_summary(date_range)

    data_frame.to_csv(query_name + ".csv")

if __name__ == "__main__":
    Main()



import setuptools

# Load the long_description from README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='mmwebreportapi',
    version='0.0.4',
    author='Manuel Huertas Lopez',
    author_email='manuel.huertas.lopez@gmail.com',
    description='Monitor Manager Web Rerpot API',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mahulo2009/monitormanager-webreportapi-python.git",
    packages=['mmwebreport', 'mmwebreport.core', 'mmwebreport.retrieve','mmwebreport.main'],
    install_requires=['pandas', 'requests'],
    entry_points={
        'console_scripts': [
            'mmreport = mmwebreport.main.main:Main'
            ],
        },
    download_url="https://github.com/mahulo2009/monitormanager-webreportapi-python/archive/refs/tags/v1.0.3-alpha.zip"
)


import json
import os
from os.path import exists

import pandas as pd

_default_extension = 'csv.gz'

class Cache(object):
    def __init__(self, root_path, date_ini, date_end, query_name, a_query):
        self._a_id = a_query[0]['component'] + "." + a_query[0]['monitor']
        self._date_ini = date_ini
        self._date_end = date_end
        self._root_path = root_path
        self._query_name = query_name
        self._a_query = a_query


    def _make_path_root(self):
        path = self._root_path + "/" + \
               self._date_ini.strftime("%Y-%m-%d") + "/" + self._date_ini.strftime("%H")
        return path

    # todo mark as abstract method
    def make_path(self, page=None, extension=_default_extension):
        pass

    def _build_query(self, page=None):
        q = {
            "date_ini": self._date_ini.strftime("%Y-%m-%d %H_%M_%S"),
            "date_end": self._date_end.strftime("%Y-%m-%d %H_%M_%S"),
            "query_name": self._query_name,
            "query": self._a_query
        }
        if page is not None:
            q["page"] = page

        return q

    def compare_query(self, query_1, query_2, index=0):
        equal = True

        equal = equal and (query_1["date_ini"] == query_2["date_ini"])
        equal = equal and (query_1["date_end"] == query_2["date_end"])

        if len(query_1["query"]) != len(query_2["query"]):
            return False

        for index in range(0, len(query_1["query"])):
            equal = equal and (query_1["query"][index]["component"] == query_2["query"][index]["component"])
            equal = equal and (query_1["query"][index]["monitor"] == query_2["query"][index]["monitor"])
            equal = equal and (query_1["query"][index]["type"] == query_2["query"][index]["type"])
        # todo review this
        #    if q1["page"] and q2["page"]:
        #        equal = equal and (q1["page"] == q2["page"])
        return equal

    def _store_query(self, file_name):
        query = {
            "date_ini": self._date_ini.strftime("%Y-%m-%d %H_%M_%S"),
            "date_end": self._date_end.strftime("%Y-%m-%d %H_%M_%S"),
            "query_name": self._query_name,
            "query": self._a_query
        }
        if not exists(os.path.dirname(file_name)):
            os.makedirs(os.path.dirname(file_name))
        with open(file_name, 'w') as outfile:
            outfile.write(json.dumps(query))

    def store_query(self, page=''):
        file_name = self.make_path(page, extension='json')
        self._store_query(file_name)

    def read_query(self, index=0, page=''):
        _id = self._a_query[index]['component'] + "." + self._a_query[index]['monitor']
        file_name = self.make_path(page, extension='json')

        with open(file_name) as json_file:
            return json.load(json_file)

    def is_data_frame_cached(self, page=''):
        file_name_cached = self.make_path(page, extension='json')

        if not exists(file_name_cached):
            return False

        query_build = self._build_query(page)
        query_disk = self.read_query(0, page)

        return self.compare_query(query_build, query_disk)

    def data_frame_from_cache(self, page=''):
        file_name = self.make_path(page)
        return pd.read_csv(file_name, compression='infer', parse_dates=[0])

    def data_frame_to_cache(self, data_frame, page=''):
        file_name = self.make_path(page)
        if not exists(os.path.dirname(file_name)):
            os.makedirs(os.path.dirname(file_name))

        data_frame.to_csv(file_name, index=False, compression='infer')
        self.store_query(page)


class CacheRaw(Cache):
    def __init__(self, root_path, date_ini, date_end, query_name, a_query):
        super().__init__(root_path, date_ini, date_end, query_name, a_query)

    def make_path(self, page=None, extension=_default_extension):
        path = self._make_path_root() + "/" + str(self._a_id).replace(".", "/") + "/raw"

        file_name_monitor = self._date_ini.strftime("%Y-%m-%d_%H_%M_%S") + \
                            "." + \
                            self._date_end.strftime("%Y-%m-%d_%H_%M_%S") + \
                            "." + \
                            str(self._a_id)
        if page is None:
            file_name_monitor = file_name_monitor + "." + self._extension
        else:
            file_name_monitor = file_name_monitor + ".raw." + str(page).rjust(4, '0') + "." + extension

        return path + "/" + file_name_monitor

    def compare_query(self, query_1, query_2, index=0):
        return super().compare_query(query_1, query_2)


class CacheFiltered(Cache):
    def __init__(self, root_path, date_ini, date_end, query_name, a_query):
        super().__init__(root_path, date_ini, date_end, query_name, a_query)

    def make_path(self, page=None, extension=_default_extension):
        path = self._make_path_root() + "/" + str(self._a_id).replace(".", "/")

        file_name_monitor = self._date_ini.strftime("%Y-%m-%d_%H_%M_%S") + \
                            "." + \
                            self._date_end.strftime("%Y-%m-%d_%H_%M_%S") + \
                            "." + \
                            str(self._a_id)

        file_name_monitor = file_name_monitor + "." + extension

        return path + "/" + file_name_monitor

    def compare_query(self, query_1, query_2, index=0):
        equal = super().compare_query(query_1, query_2)

        for index in range(0, len(query_1["query"])):
            equal = equal and (query_1["query"][index]["epsilon"] == query_2["query"][index]["epsilon"])

        return equal


class CacheSummaryHourly(Cache):
    def __init__(self, root_path, date_ini, date_end, query_name, a_query):
        super().__init__(root_path, date_ini, date_end, query_name, a_query)

    def make_path(self, page=None, extension='cvs.gz'):
        path = self._make_path_root() + "/summary"

        file_name_monitor = self._date_ini.strftime("%Y-%m-%d_%H_%M_%S") + \
                            "." + \
                            self._date_end.strftime("%Y-%m-%d_%H_%M_%S")

        file_name_monitor = file_name_monitor + "." + self._query_name + "." + "summary" + "." + extension

        return path + "/" + file_name_monitor

    def compare_query(self, query_1, query_2, index=0):
        equal = super().compare_query(query_1, query_2)

        for index in range(0, len(query_1["query"])):
            equal = equal and (query_1["query"][index]["epsilon"] == query_2["query"][index]["epsilon"])

        return equal


class CacheSummary(Cache):
    def __init__(self, root_path, date_ini, date_end, query_name, a_query):
        super().__init__(root_path, date_ini, date_end, query_name, a_query)

    def make_path(self, page=None, extension='cvs.gz'):
        path = self._root_path + "/summary/" + self._query_name

        file_name_monitor = self._date_ini.strftime("%Y-%m-%d_%H_%M_%S") + \
                            "." + \
                            self._date_end.strftime("%Y-%m-%d_%H_%M_%S")

        file_name_monitor = file_name_monitor + "." + self._query_name + "." + "summary" + "." + extension

        return path + "/" + file_name_monitor

    def compare_query(self, query_1, query_2, index=0):
        equal = super().compare_query(query_1, query_2)

        for index in range(0, len(query_1["query"])):
            equal = equal and (query_1["query"][index]["epsilon"] == query_2["query"][index]["epsilon"])

        return equal
